'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _class, _temp;

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _react = require('react');

var _react2 = _interopRequireDefault(_react);

var _propTypes = require('prop-types');

var _propTypes2 = _interopRequireDefault(_propTypes);

var _immutable = require('immutable');

var _isoFetchStream = require('iso-fetch-stream');

var _isoFetchStream2 = _interopRequireDefault(_isoFetchStream);

var _schema = require('./lib/schema');

var _store = require('./store');

var _store2 = _interopRequireDefault(_store);

var _log = require('./lib/log');

var _log2 = _interopRequireDefault(_log);

var _query_string = require('./lib/query_string');

var _query_string2 = _interopRequireDefault(_query_string);

var _config = require('./config');

var _config2 = _interopRequireDefault(_config);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _objectWithoutProperties(obj, keys) { var target = {}; for (var i in obj) { if (keys.indexOf(i) >= 0) continue; if (!Object.prototype.hasOwnProperty.call(obj, i)) continue; target[i] = obj[i]; } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

/* basic state tree design for api.js
* {
*   $api: {
*     [url]: {
*       [resource]: {
*         models: {
*           [id]: {
*             [key]: any,
*             _fetched?: boolean
*           }
*         },
*         requests: {
*           [url]: {
*             pending?: boolean,
*             error?: Object,
*             data?: List | Map
*           }
*         }
*       }
*     }
*   }
* }
*/

// TODO: use some more specific method from store's undo implementation
var isUndo = function isUndo() {
  return _store2.default.getState('_reset');
};

var methodDict = {
  create: 'post',
  list: 'get',
  get: 'get',
  update: 'put',
  delete: 'delete',
  search: 'get'
};

var logRoutes = function logRoutes(routes, resource, ns) {
  if (_config2.default.mode === 'development') {
    /* eslint-disable no-console */
    console.groupCollapsed('Rendering Api component for %c' + resource, 'color: green', 'with these routes');
    console.table(Object.keys(routes).reduce(function (memo, key) {
      var val = routes[key]('id', {});
      var def = typeof val === 'string' ? { method: methodDict[key], route: val } : val;
      var arg = key === 'get' ? 'id' : '';
      return Object.assign(memo, _defineProperty({}, key, Object.assign(def, _defineProperty({}, 'props.' + ns + '.<fn>', '$' + key + '(' + arg + ')'))));
    }, {}));
    console.groupEnd();
  }
};

function createActions(_ref) {
  var url = _ref.url,
      props = _objectWithoutProperties(_ref, ['url']);

  var fetchOptions = props.credentials ? { credentials: props.credentials } : {};
  if (props.auth || props.authorization) fetchOptions.headers = { Authorization: props.auth || props.authorization };
  var fetch = (0, _isoFetchStream2.default)(fetchOptions);

  return Object.keys(props).reduce(function (memo, key) {

    if ((0, _schema.isSchema)(props[key])) {

      var schema = props[key].schema || props[key];
      var getData = props[key].getData || function (data) {
        return data;
      };
      var idField = (0, _schema.getIdField)(schema);
      var resourceType = schema.title;
      var resourcePath = '/' + resourceType;

      var routes = Object.assign({
        create: function create() {
          return { method: 'post', route: resourcePath };
        },
        list: function list() {
          return { method: 'get', route: resourcePath };
        },
        search: function search() {
          return { method: 'get', route: resourcePath };
        },
        get: function get(id) {
          return { method: 'get', route: resourcePath + '/' + id };
        },
        update: function update(id) {
          return { method: 'put', route: resourcePath + '/' + id };
        },
        delete: function _delete(id) {
          return { method: 'delete', route: resourcePath + '/' + id };
        }
      }, props[key].routes || {});

      logRoutes(routes, resourceType, key);

      var getState = function getState(key) {
        return _store2.default.getState(['$api', url, resourceType].concat(key || []).map(function (item) {
          return String(item);
        }));
      };
      var setState = function setState(val, key) {
        return _store2.default.setState(['$api', url, resourceType].concat(key || []).map(function (item) {
          return String(item);
        }), val);
      };
      var setStateQuiet = function setStateQuiet(val, key) {
        return _store2.default.setStateQuiet(['$api', url, resourceType].concat(key || []), val);
      };

      var getRequests = function getRequests(path) {
        return getState(['requests'].concat(path || []));
      };
      var getRequestsData = function getRequestsData(path) {
        return getRequests([path, 'data']);
      };
      var setRequests = function setRequests(data, path) {
        return setState(data, ['requests'].concat(path || []));
      };
      var setRequestsData = function setRequestsData(path, data) {
        return setState(data, ['requests', path, 'data']);
      };

      var getCollection = function getCollection(path) {
        var collection = getRequestsData(path);
        if (collection) {
          var models = getModels();
          return collection.reduce(function (memo, id) {
            var model = models.get(id);
            if (model) return memo.push(model);
            return memo;
          }, (0, _immutable.List)());
        } else {
          return null;
        }
      };

      var setCollection = function setCollection() {
        var data = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : (0, _immutable.List)();
        var path = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : '/';

        var state = getState();
        var nextState = state.withMutations(function (map) {
          var dict = data.reduce(function (memo, item) {
            return _extends({}, memo, _defineProperty({}, item[idField], item));
          }, {});
          map.set('models', getModels().mergeDeep(dict));
          map.setIn(['requests', path, 'data'], (0, _immutable.List)(Object.keys(dict)));
        });
        setState(nextState);
      };

      var clearCollection = function clearCollection() {
        var path = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : '/';
        return setRequestsData(path, null);
      };
      var clearAll = function clearAll() {
        setRequests((0, _immutable.Map)());
        setModels((0, _immutable.Map)());
      };

      var getModels = function getModels() {
        return getState('models') || (0, _immutable.Map)();
      };
      var setModels = function setModels(data) {
        return setState(data, 'models');
      };
      var getModel = function getModel(id) {
        return getState(['models', id]);
      };
      var setModel = function setModel(id, data) {
        return setState(data, ['models', id]);
      };
      var getPending = function getPending(path) {
        return getState(['requests', path, 'pending']);
      };
      var setPending = function setPending(path, data) {
        return setStateQuiet(data, ['requests', path, 'pending']);
      };
      var getError = function getError(path) {
        return getState(['requests', path, 'error']);
      };
      var setError = function setError(path, error) {
        var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

        var method = options.quiet ? setStateQuiet : setState;
        method(error, ['requests', path, 'error']);
      };

      var setSearchResults = function setSearchResults(path, data) {
        return setCollection(data, path);
      };
      var getSearchResults = function getSearchResults(path) {
        return getCollection(path);
      };

      var removeFromCollections = function removeFromCollections(map, id) {
        return [].concat(_toConsumableArray(map.get('requests').entries())).reduce(function (undo, _ref2) {
          var _ref3 = _slicedToArray(_ref2, 2),
              path = _ref3[0],
              request = _ref3[1];

          var collection = request.get('data');
          if (_immutable.List.isList(collection)) {
            var modelIdx = collection.findIndex(function (modelId) {
              return modelId === id;
            });
            if (~modelIdx) {
              var nextCollection = collection.delete(modelIdx);
              map.setIn(['requests', path, 'data'], nextCollection);
              undo.push(function () {
                return setCollection(nextCollection.insert(modelIdx, id), path);
              });
            }
          }
          return undo;
        }, []);
      };

      var deleteModel = function deleteModel(id) {
        var undo = [];
        var state = getState();
        setState(state.withMutations(function (map) {
          var model = getModel(id);
          if (model) {
            map.set('models', getModels().delete(id));
            undo.push(function () {
              return setModels(getModels().set(id, model));
            });
            undo.push.apply(undo, _toConsumableArray(removeFromCollections(map, id)));
          }
        }));
        return undo;
      };

      var updateModel = function updateModel(id, vals) {
        var undo = [];
        var model = getModel(id);
        if (model) {
          setModel(id, model.mergeDeep(vals));
          undo.push(function () {
            return setModels(getModels().set(id, model));
          });
        }
        return undo;
      };

      var shouldFetch = function shouldFetch(path) {
        return !isUndo() && !getPending(path) && !getError(path);
      };

      var api = ['get', 'post', 'put', 'delete'].reduce(function (memo, method) {
        return _extends({}, memo, _defineProperty({}, method, function (path) {
          for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
            args[_key - 1] = arguments[_key];
          }

          setPending(path, true);
          return fetch[method].apply(fetch, ['' + url + path].concat(args)).then(function (response) {
            var data = getData(response);
            setPending(path, false);
            setError(path, null, { quiet: true });
            return data;
          }, function (err) {
            setPending(path, false);
            setError(path, err);
            return Promise.reject(err);
          });
        }));
      }, {});

      var getRouteConfig = function getRouteConfig(config) {
        var defaultMethod = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 'get';
        return typeof config === 'string' ? { method: defaultMethod, route: config } : { method: config.method.toLowerCase(), route: config.route };
      };

      /**
       * api calls
       */
      var fetchAll = function fetchAll(path) {
        var _getRouteConfig = getRouteConfig(routes.list()),
            defaultRoute = _getRouteConfig.route,
            method = _getRouteConfig.method;

        var route = path || defaultRoute;
        if (shouldFetch(route)) {
          api[method](route).then(function (data) {
            return setCollection(data, route);
          });
        }
      };

      var fetchOne = function fetchOne(id) {
        var _getRouteConfig2 = getRouteConfig(routes.get(id)),
            route = _getRouteConfig2.route,
            method = _getRouteConfig2.method;

        if (shouldFetch(route)) {
          api[method](route).then(function (data) {
            return setModel(data[idField], _extends({}, data, { _fetched: true }));
          });
        }
      };

      var createOne = function createOne(data) {
        var _getRouteConfig3 = getRouteConfig(routes.create(data), 'post'),
            route = _getRouteConfig3.route,
            method = _getRouteConfig3.method;

        return api[method](route, data).then(function (data) {
          fetchAll();
          return data;
        });
      };

      var updateOne = function updateOne(id, data) {
        var _getRouteConfig4 = getRouteConfig(routes.update(id, data), 'put'),
            route = _getRouteConfig4.route,
            method = _getRouteConfig4.method;

        return api[method](route, data);
      };

      var deleteOne = function deleteOne(id) {
        var _getRouteConfig5 = getRouteConfig(routes.delete(id), 'delete'),
            route = _getRouteConfig5.route,
            method = _getRouteConfig5.method;

        return api[method](route);
      };

      /**/

      var $clear = function $clear(id) {
        return function () {
          return setModel(id, null);
        };
      };
      var $reset = function $reset(id) {
        return function () {
          return fetchOne(id);
        };
      };

      var optimisticDelete = function optimisticDelete(id) {
        var undoDelete = deleteModel(id);
        if (undoDelete.length) {
          return deleteOne(id).catch(function (err) {
            (0, _log.logError)('Failed to delete ' + id, err);
            undoDelete.forEach(function (undo) {
              return undo();
            });
            return Promise.reject(err);
          });
        } else {
          return Promise.reject(new Error(404));
        }
      };

      var $delete = function $delete(id) {
        return function () {
          var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
          return options.optimistic === false || !getModel(id) ? deleteOne(id).then(function () {
            return deleteModel(id);
          }) : optimisticDelete(id);
        };
      };

      var optimisticUpdate = function optimisticUpdate(id, vals) {
        var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

        var placeholder = typeof options.optimistic === 'function' ? options.optimistic(_extends({ id: id }, vals)) : vals;
        var undoUpdate = updateModel(id, placeholder);
        if (undoUpdate.length) {
          return updateOne(id, vals).catch(function (err) {
            (0, _log.logError)('Failed to update ' + id + ' with vals', vals, err);
            undoUpdate.forEach(function (undo) {
              return undo();
            });
            return Promise.reject(err);
          });
        } else {
          return Promise.reject(new Error(404));
        }
      };

      var $update = function $update(id) {
        return function (vals) {
          var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
          return options.optimistic === false || !getModel(id) ? updateOne(id, vals).then(function (data) {
            return updateModel(id, data);
          }) : optimisticUpdate(id, vals, options);
        };
      };

      var getPlaceholder = function getPlaceholder() {
        var path = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;
        var dataType = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : _immutable.List;

        var placeholder = dataType();
        placeholder.$isPending = true;
        placeholder.$error = path ? getError(path) : null;
        return placeholder;
      };

      var addRestMethods = function addRestMethods(model) {
        var id = model.get(idField);
        model.$delete = $delete(id);
        model.$update = $update(id);
        model.$clear = $clear(id);
        model.$reset = $reset(id);
        return model;
      };

      var $list = function $list(params) {
        var path = routes.list(params) + (params ? '?' + (0, _query_string2.default)(params) : '');
        var collection = getCollection(path);
        if (!collection) {
          fetchAll(path);
          return getPlaceholder(path);
        } else {
          return collection.map(addRestMethods);
        }
      };

      var main = function main(params) {
        return $list(params);
      };

      main.$list = $list;

      main.$get = function (id) {
        var model = getModel(id);
        if (!model || !model.get('_fetched')) {
          var path = routes.get(id);
          fetchOne(id);
          var placeholder = getPlaceholder(path, _immutable.Map);
          placeholder.$delete = $delete(id);
          placeholder.$update = $update(id);
          return placeholder;
        } else {
          return addRestMethods(model);
        }
      };

      // remove the cache for the resource collection
      main.$clear = function () {
        return clearCollection();
      };
      main.$clearAll = clearAll;
      main.$reset = fetchAll;

      var optimisticCreate = function optimisticCreate(data) {
        var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

        if (typeof options.optimistic === 'function') {
          var nextState = getState().withMutations(function (map) {
            return options.optimistic(map, data);
          });
          if (nextState) setState(nextState);
          return createOne(data);
        } else {
          (0, _log.logWarn)('Optimistic creates must receive a function that updates the state with the optimistic data. The create will proceed pessimistically.');
          return createOne(data);
        }
      };

      main.$create = function (data) {
        var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
        return !options.optimistic ? createOne(data) : optimisticCreate(data, options);
      };

      main.$search = function (params) {
        var _getRouteConfig6 = getRouteConfig(routes.search(params)),
            method = _getRouteConfig6.method,
            route = _getRouteConfig6.route;

        var fullRoute = method === 'get' ? route + ('?' + (0, _query_string2.default)(params)) : route;
        if (shouldFetch(fullRoute)) {
          var promise = method === 'get' ? api.get(fullRoute) : api[method](fullRoute, params);

          return promise.then(function (data) {
            setSearchResults(fullRoute, data);
            return data;
          });
        }
        // TODO: store promise as pending value so it can be used on repeat
        // calls
        return Promise.resolve('pending');
      };

      main.$search.results = function (params) {
        var _getRouteConfig7 = getRouteConfig(routes.search(params)),
            route = _getRouteConfig7.route,
            method = _getRouteConfig7.method;

        var fullRoute = method === 'get' ? route + ('?' + (0, _query_string2.default)(params)) : route;
        var resultsCached = getSearchResults(fullRoute);
        if (resultsCached) {
          return addRestMethods(resultsCached);
        } else {
          var placeholder = (0, _immutable.List)();
          placeholder.$isPending = !!getPending(fullRoute);
          placeholder.$error = getError(fullRoute);
          return placeholder;
        }
      };

      // hang standard api methods off of .api so devs can construct
      // non-standard paths
      main.api = ['delete', '$get', 'get', 'post', 'put', 'stream'].reduce(function (memo, method) {
        return _extends({}, memo, _defineProperty({}, method, function (path) {
          for (var _len2 = arguments.length, args = Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
            args[_key2 - 1] = arguments[_key2];
          }

          var route = resourcePath + path;
          if (method === '$get') {
            var cache = getRequestsData(route);
            if (cache) {
              return _immutable.List.isList(cache) ? getCollection(route).map(addRestMethods) : cache;
            } else {
              if (shouldFetch(route)) {
                api.get.apply(api, [route].concat(args)).then(function (data) {
                  if (Array.isArray(data)) {
                    setCollection(data, route);
                  } else {
                    setRequestsData(route, data);
                  }
                  return data;
                });
              }
              return getPlaceholder(route);
            }
          } else {
            return api[method].apply(api, [route].concat(args));
          }
        }));
      }, {});

      main._schema = schema;
      main._resourceType = resourceType;
      main.getState = getState;

      memo[key] = main;
      if (typeof window !== 'undefined') {
        window.jetset = window.jetset || {};
        window.jetset[key] = main;
      }
    }
    return memo;
  }, {});
}

var Api = (_temp = _class = function (_React$Component) {
  _inherits(Api, _React$Component);

  function Api(props) {
    _classCallCheck(this, Api);

    var _this = _possibleConstructorReturn(this, (Api.__proto__ || Object.getPrototypeOf(Api)).call(this, props));

    _this.subscriptions = [];

    _this.api = createActions(props);
    _this.state = { cache: null };
    return _this;
  }

  _createClass(Api, [{
    key: 'componentWillMount',
    value: function componentWillMount() {
      var _this2 = this;

      this.subscriptions = Object.keys(this.api).map(function (key) {
        var resource = _this2.api[key]._resourceType;
        return _store2.default.subscribeTo(['$api', _this2.props.url, resource], function (state) {
          (0, _log2.default)('\uD83C\uDF00 <Api> is re-rendering based on state changes on branch: %c' + (_this2.props.url + ' ‣ ' + resource), 'color: #5B4532');
          _this2.setState({ cache: state });
        });
      });
    }
  }, {
    key: 'componentWillUnmount',
    value: function componentWillUnmount() {
      this.subscriptions.forEach(function (subscription) {
        return _store2.default.unsubscribe(subscription);
      });
    }
  }, {
    key: 'render',
    value: function render() {
      var _props = this.props,
          children = _props.children,
          props = _objectWithoutProperties(_props, ['children']);

      return children && typeof children.type === 'function' ? _react2.default.cloneElement(children, _extends({}, props, this.api)) : children;
    }
  }]);

  return Api;
}(_react2.default.Component), _class.propTypes = {
  url: _propTypes2.default.string.isRequired,
  // see https://github.com/github/fetch#sending-cookies for reference
  credentials: _propTypes2.default.oneOf(['same-origin', 'include']),
  token: _propTypes2.default.string
}, _temp);
exports.default = Api;